#    Copyright (C) 2014 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""Widget to plot 3D axes, and to handle conversion of coordinates to plot
positions."""

import math
import numpy as N
import itertools

from . import widget
from . import axisticks
from . import axis
from .. import qtall as qt
from .. import document
from .. import setting
from .. import utils
from ..helpers import threed

def _(text, disambiguation=None, context='Axis3D'):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

class _AxisLabels(threed.AxisLabels):
    """For drawing tick labels.

    box1,2: points at corners of graph
    tickfracs: fractions along axis for tick labels
    ticklabels: list of labels to plot at fracs
    ticklabelsprop: font properties for ticks
    axislabel: label for axis
    axislabelposn: position of axis label (or -1 for no label)
    axislabelprop: font properties for axislabel
    """

    def __init__(self, box1, box2, tickfracs, ticklabels, ticklabelsprop,
                 axislabel, axislabelposn, axislabelprop):
        threed.AxisLabels.__init__(self, box1, box2, tickfracs, axislabelposn)
        self.ticklabels = ticklabels
        self.ticklabelsprop = ticklabelsprop
        self.axislabel = axislabel
        self.axislabelposn = axislabelposn
        self.axislabelprop = axislabelprop

    def drawLabel(self, painter, index, pt, ax1, ax2, axangle):
        """Called when scene wants to paint label."""

        painter.translate(pt.x(), pt.y())

        # angle of axis inclination
        angle = math.atan2(ax2.y()-ax1.y(), ax2.x()-ax1.x()) * (180/math.pi)

        # vector to axis from graph centre
        cvecy = math.sin(axangle*math.pi/180)
        cvecx = math.cos(axangle*math.pi/180)

        # character upright vector
        avecy = math.sin((angle-90)*math.pi/180)
        avecx = math.cos((angle-90)*math.pi/180)
        dot = cvecx*avecx+cvecy*avecy

        #print('axangle % 7.2f  angle % 7.2f  delta % 7.2f  dot % 7.2f' % (
        #    axangle, angle, axangle+angle, dot), self.ticklabels[index])

        # flip depending on relative direction of label and graph centre
        if dot < 0:
            angle = angle+180
            if angle > 180:
                angle = angle-360

        # flip if upside down
        if angle < -90 or angle > 90:
            angle = angle+180
            valign = 1
        else:
            valign = -1

        painter.rotate(angle)

        if index >= 0:
            self.drawTickLabel(painter, pt.x(), pt.y(), angle, index, valign)
        else:
            self.drawAxisLabel(painter, valign)

    def drawTickLabel(self, painter, x, y, angle, index, valign):
        """Draw a tick label."""

        font = self.ticklabelsprop.makeQFont(painter)
        painter.setFont(font)

        label = self.ticklabels[index]
        renderer = utils.Renderer(
            painter, font, 0, 0, label,
            alignhorz=0, alignvert=valign,
            usefullheight=True)

        # get text bounds
        rect = renderer.getTightBounds()
        rect.rotateAboutOrigin(angle * math.pi/180)
        rect.translate(x, y)

        # draw text if it doesn't overlap with existing text
        if not painter.textrects.willOverlap(rect):
            pen = self.ticklabelsprop.makeQPen(painter)
            painter.setPen(pen)
            renderer.render()
            painter.textrects.addRect(rect)

    def drawAxisLabel(self, painter, valign):
        """Draw label for axis."""

        # y increment from labels closer to axis
        deltay = 0
        if self.ticklabels:
            font = self.ticklabelsprop.makeQFont(painter)
            fm = utils.FontMetrics(font, painter.device())
            deltay = valign*fm.lineSpacing()

        # change alignment depending on label position
        if self.axislabelposn==0.:
            halign = -1
        elif self.axislabelposn==1.:
            halign = 1
        else:
            halign = 0

        # draw label
        font = self.axislabelprop.makeQFont(painter)
        painter.setFont(font)
        pen = self.axislabelprop.makeQPen(painter)
        painter.setPen(pen)

        renderer = utils.Renderer(
            painter, font, 0, deltay, self.axislabel,
            alignhorz=halign, alignvert=valign,
            usefullheight=True)
        renderer.render()

class MajorTick(setting.Line3D):
    '''Major tick settings.'''

    def __init__(self, name, **args):
        setting.Line3D.__init__(self, name, **args)
        self.get('color').newDefault('grey')
        self.add(setting.Float(
            'length',
            20.,
            descr = _('Length of major ticks'),
            usertext= _('Length')))
        self.add(setting.Int(
            'number',
            6,
            descr = _('Number of major ticks to aim for'),
            usertext= _('Number')))
        self.add(setting.FloatList(
            'manualTicks',
            [],
            descr = _('List of tick values overriding defaults'),
            usertext= _('Manual ticks')))

class MinorTick(setting.Line3D):
    '''Minor tick settings.'''

    def __init__(self, name, **args):
        setting.Line3D.__init__(self, name, **args)
        self.get('color').newDefault('grey')
        self.add( setting.Float(
            'length',
            10,
            descr = _('Length of minor ticks'),
            usertext= _('Length')))
        self.add( setting.Int(
            'number',
            20,
            descr = _('Number of minor ticks to aim for'),
            usertext= _('Number')))

class GridLine(setting.Line3D):
    '''Grid line settings.'''

    def __init__(self, name, **args):
        setting.Line3D.__init__(self, name, **args)

        self.get('color').newDefault('grey')
        self.get('hide').newDefault(True)

class MinorGridLine(setting.Line3D):
    '''Minor tick grid line settings.'''

    def __init__(self, name, **args):
        setting.Line3D.__init__(self, name, **args)

        self.get('color').newDefault('lightgrey')
        self.get('hide').newDefault(True)

class AxisLabel(setting.Text):
    """Axis label."""

    def __init__(self, name, **args):
        setting.Text.__init__(self, name, **args)
        self.add( setting.Choice(
            'position',
            ('at-minimum', 'centre', 'at-maximum'),
            'centre',
            descr = _('Position of axis label'),
            usertext = _('Position') ) )

class TickLabel(axis.TickLabel):
    """3D axis tick label."""
    def __init__(self, name, **args):
        axis.TickLabel.__init__(self, name, **args)
        self.remove('rotate')
        self.remove('offset')

class Axis3D(widget.Widget):
    """Manages and draws an axis."""

    typename = 'axis3d'
    allowusercreation = True
    description = _('3D axis')
    isaxis = True
    isaxis3d = True

    def __init__(self, parent, name=None):
        """Initialise axis."""

        widget.Widget.__init__(self, parent, name=name)
        s = self.settings

        for n in ('x', 'y', 'z'):
            if self.name == n and s.direction != n:
                s.direction = n

        # automatic range
        self.setAutoRange(None)

        # document updates change set variable when things need recalculating
        self.docchangeset = -1

    @classmethod
    def addSettings(klass, s):
        """Construct list of settings."""
        widget.Widget.addSettings(s)
        s.add( setting.Str(
            'label', '',
            descr=_('Axis label text'),
            usertext=_('Label')) )
        s.add( setting.AxisBound(
            'min', 'Auto',
            descr=_('Minimum value of axis'),
            usertext=_('Min')) )
        s.add( setting.AxisBound(
            'max', 'Auto',
            descr=_('Maximum value of axis'),
            usertext=_('Max')) )
        s.add( setting.Bool(
            'log', False,
            descr=_('Whether axis is logarithmic'),
            usertext=_('Log')) )
        s.add( axis.AutoRange(
            'autoRange', 'next-tick') )
        s.add( setting.Choice(
            'mode',
            ('numeric', 'datetime', 'labels'),
            'numeric',
            descr=_('Type of ticks to show on on axis'),
            usertext=_('Mode')) )

        s.add( setting.Bool(
            'autoMirror', True,
            descr=_('Place axis on opposite side of graph if none'),
            usertext=_('Auto mirror'),
            formatting=True) )

        s.add( setting.Float(
            'datascale', 1.,
            descr=_('Scale data plotted by this factor'),
            usertext=_('Scale')) )

        s.add( setting.Choice(
            'direction',
            ['x', 'y', 'z'],
            'x',
            descr=_('Direction of axis'),
            usertext=_('Direction')) )
        s.add( setting.Float(
            'lowerPosition', 0.,
            descr=_('Fractional position of lower end of axis on graph'),
            usertext=_('Min position')) )
        s.add( setting.Float(
            'upperPosition', 1.,
            descr=_('Fractional position of upper end of axis on graph'),
            usertext=_('Max position')) )
        s.add( setting.Float(
            'otherPosition1', 0.,
            descr=_(
                'Fractional position of axis in its perpendicular direction 1'),
            usertext=_('Axis position 1')) )
        s.add( setting.Float(
            'otherPosition2', 0.,
            descr=_(
                'Fractional position of axis in its perpendicular direction 2'),
            usertext=_('Axis position 2')) )

        s.add( setting.Line3D(
            'Line',
            descr=_('Axis line settings'),
            usertext=_('Axis line')), pixmap='settings_axisline' )
        s.add( AxisLabel(
            'Label',
            descr=_('Axis label settings'),
            usertext=_('Axis label')), pixmap='settings_axislabel' )
        s.add( TickLabel(
            'TickLabels',
            descr=_('Tick label settings'),
            usertext=_('Tick labels')), pixmap='settings_axisticklabels' )
        s.add( MajorTick(
            'MajorTicks',
            descr=_('Major tick line settings'),
            usertext=_('Major ticks')), pixmap='settings_axismajorticks' )
        s.add( MinorTick(
            'MinorTicks',
            descr=_('Minor tick line settings'),
            usertext=_('Minor ticks')),  pixmap='settings_axisminorticks' )
        s.add( GridLine(
            'GridLines',
            descr=_('Grid line settings'),
            usertext=_('Grid lines')), pixmap='settings_axisgridlines' )
        s.add( MinorGridLine(
            'MinorGridLines',
            descr=_('Minor grid line settings'),
            usertext=_('Grid lines for minor ticks')),
            pixmap='settings_axisminorgridlines' )

    @classmethod
    def allowedParentTypes(self):
        from . import graph3d
        return (graph3d.Graph3D,)

    @property
    def userdescription(self):
        """User friendly description."""
        s = self.settings
        return "range %s to %s%s" % (
            str(s.min), str(s.max), ['',' (log)'][s.log])

    def isLinked(self):
        """Whether is an axis linked to another."""
        return False

    def setAutoRange(self, autorange):
        """Set the automatic range of this axis (called from page helper)."""

        if autorange:
            scale = self.settings.datascale
            self.autorange = ar = [x*scale for x in autorange]
            if self.settings.log:
                ar[0] = max(1e-99, ar[0])
        else:
            if self.settings.log:
                self.autorange = [1e-2, 1.]
            else:
                self.autorange = [0., 1.]

    def usesAutoRange(self):
        """Return whether any of the bounds are automatically determined."""
        return self.settings.min == 'Auto' or self.settings.max == 'Auto'

    def computePlottedRange(self, force=False, overriderange=None):
        """Convert the range requested into a plotted range."""

        if self.docchangeset == self.document.changeset and not force:
            return

        s = self.settings
        if overriderange is None:
            self.plottedrange = [s.min, s.max]
        else:
            self.plottedrange = overriderange

        # automatic lookup of minimum
        if overriderange is None:
            if s.min == 'Auto':
                self.plottedrange[0] = self.autorange[0]
            if s.max == 'Auto':
                self.plottedrange[1] = self.autorange[1]

        # yuck, but sometimes it's true
        # tweak range to make sure things don't blow up further down the
        # line
        if ( abs(self.plottedrange[0] - self.plottedrange[1]) <
             ( abs(self.plottedrange[0]) + abs(self.plottedrange[1]) )*1e-8 ):
            self.plottedrange[1] = ( self.plottedrange[0] +
                                     max(1., self.plottedrange[0]*0.1) )

        # handle axis values round the wrong way
        invertaxis = self.plottedrange[0] > self.plottedrange[1]
        if invertaxis:
            self.plottedrange = self.plottedrange[::-1]

        # make sure log axes don't blow up
        if s.log:
            if self.plottedrange[0] < 1e-99:
                self.plottedrange[0] = 1e-99
            if self.plottedrange[1] < 1e-99:
                self.plottedrange[1] = 1e-99
            if self.plottedrange[0] == self.plottedrange[1]:
                self.plottedrange[1] = self.plottedrange[0]*2

        s.get('autoRange').adjustPlottedRange(
            self.plottedrange, s.min=='Auto', s.max=='Auto', s.log, self.document)

        self.computeTicks()

        # invert bounds if axis was inverted
        if invertaxis:
            self.plottedrange = self.plottedrange[::-1]

        self.docchangeset = self.document.changeset

    def computeTicks(self, allowauto=True):
        """Update ticks given plotted range.
        if allowauto is False, then do not allow ticks to be
        updated
        """

        s = self.settings

        if s.mode in ('numeric', 'labels'):
            tickclass = axisticks.AxisTicks
        else:
            tickclass = axisticks.DateTicks

        nexttick = s.autoRange == 'next-tick'
        extendmin = nexttick and s.min == 'Auto' and allowauto
        extendmax = nexttick and s.max == 'Auto' and allowauto

        # create object to compute ticks
        axs = tickclass(
            self.plottedrange[0], self.plottedrange[1],
            s.MajorTicks.number, s.MinorTicks.number,
            extendmin = extendmin, extendmax = extendmax,
            logaxis = s.log
        )

        axs.getTicks()
        self.plottedrange[0] = axs.minval
        self.plottedrange[1] = axs.maxval
        self.majortickscalc = axs.tickvals
        self.minortickscalc = axs.minorticks
        self.autoformat = axs.autoformat

        # override values if requested
        if len(s.MajorTicks.manualTicks) > 0:
            ticks = []
            for i in s.MajorTicks.manualTicks:
                if i >= self.plottedrange[0] and i <= self.plottedrange[1]:
                    ticks.append(i)
            self.majortickscalc = N.array(ticks)

    def getPlottedRange(self):
        """Return the range plotted by the axes."""
        self.computePlottedRange()
        return (self.plottedrange[0], self.plottedrange[1])

    def dataToLogicalCoords(self, vals, scaling=True):
        """Compute coordinates on graph to logical graph coordinates (0..1)

        If scaling is True, apply scaling factor for data
        """

        self.computePlottedRange()
        s = self.settings

        svals = vals*s.datascale if scaling else vals
        if s.log:
            fracposns = self.logConvertToPlotter(svals)
        else:
            fracposns = self.linearConvertToPlotter(svals)

        lower, upper = s.lowerPosition, s.upperPosition
        return lower + fracposns*(upper-lower)

    def linearConvertToPlotter(self, v):
        """Convert graph coordinates to 0..1 coordinates"""
        return ( (v - self.plottedrange[0]) /
                 (self.plottedrange[1] - self.plottedrange[0]) )

    def logConvertToPlotter(self, v):
        """Convert graph coordinates to 0..1 coordinates"""
        log1 = N.log(self.plottedrange[0])
        log2 = N.log(self.plottedrange[1])
        return (N.log(N.clip(v, 1e-99, 1e99)) - log1) / (log2 - log1)

    def transformToAxis(self, v):
        """Return value to give equal-spaced values in transformed coordinates."""
        x = v*self.settings.datascale
        return N.log(x) if self.settings.log else x

    def transformFromAxis(self, v):
        """Convert transformed values back."""
        x = v/self.settings.datascale
        return N.exp(x) if self.settings.log else x

    def getAutoMirrorCombs(self):
        """Get combinations of other position for auto mirroring."""
        s = self.settings
        op1 = s.otherPosition1
        op2 = s.otherPosition2
        if not s.autoMirror:
            return ((op1, op2),)
        if op1 == 0 or op1 == 1:
            op1list = [1, 0]
        else:
            op1list = [op1]
        if op2 == 0 or op2 == 1:
            op2list = [1, 0]
        else:
            op2list = [op2]
        return itertools.product(op1list, op2list)

    def addAxisLine(self, painter, cont, dirn):
        """Build list of lines to draw axis line, mirroring if necessary.

        Returns list of start and end points of axis lines
        """

        s = self.settings
        lower, upper = s.lowerPosition, s.upperPosition

        outstart = []
        outend = []
        for op1, op2 in self.getAutoMirrorCombs():
            if dirn == 'x':
                outstart += [(lower, op1, op2)]
                outend += [(upper, op1, op2)]
            elif dirn == 'y':
                outstart += [(op1, lower, op2)]
                outend += [(op1, upper, op2)]
            else:
                outstart += [(op1, op2, lower)]
                outend += [(op1, op2, upper)]

        if not s.Line.hide:
            startpts = threed.ValVector(N.ravel(outstart))
            endpts = threed.ValVector(N.ravel(outend))
            lineprop = s.Line.makeLineProp(painter)
            cont.addObject(threed.LineSegments(startpts, endpts, lineprop))

        return list(zip(outstart, outend))

    def addLabels(self, cont, linecoords, ticklabelsprop, tickfracs, tickvals,
                  axislabel, axislabelprop):
        """Make tick labels for axis."""

        if not ticklabelsprop.hide:
            # make strings for labels
            fmt = ticklabelsprop.format
            if fmt.lower() == 'auto':
                fmt = self.autoformat
            scale = ticklabelsprop.scale
            ticklabels = [
                utils.formatNumber(v*scale, fmt, locale=self.document.locale)
                for v in tickvals ]
        else:
            return

        # disable drawing of label
        if axislabelprop.hide:
            axislabel = ""
            axislabelposn = -1
        else:
            axislabelposn = {
                'at-minimum': 0,
                'centre': 0.5,
                'at-maximum': 1}[axislabelprop.position]

        # this is to get the z ordering of the ends of axes correct
        # where two axes join together
        tf = N.array(tickfracs)
        tf[tf==0.] = 0.00001
        tf[tf==1.] = 0.99999
        atl = _AxisLabels(
            threed.Vec3(0,0,0), threed.Vec3(1,1,1),
            threed.ValVector(tf), ticklabels, ticklabelsprop,
            axislabel, axislabelposn, axislabelprop)

        for startpos, endpos in linecoords:
            atl.addAxisChoice(threed.Vec3(*startpos), threed.Vec3(*endpos))
        cont.addObject(atl)

    def addAxisTicks(self, painter, cont, dirn, linecoords, tickprops,
                     ticklabelsprop, tickvals):
        """Add ticks for the vals and tick properties class given.
        linecoords: coordinates of start and end points of lines
        labelprops: properties of label, or None
        cont: container to add ticks
        dirn: 'x', 'y', 'z' for axis
        """

        ticklen = tickprops.length * 1e-3
        tfracs = self.dataToLogicalCoords(tickvals, scaling=False)

        outstart = []
        outend = []
        for op1, op2 in self.getAutoMirrorCombs():
            # where to draw tick from
            op1pts = N.full_like(tfracs, op1)
            op2pts = N.full_like(tfracs, op2)
            # where to draw tick to
            op1pts2 = N.full_like(tfracs, op1+ticklen*(1 if op1 < 0.5 else -1))
            op2pts2 = N.full_like(tfracs, op2+ticklen*(1 if op2 < 0.5 else -1))

            # swap coordinates depending on axis direction
            if dirn == 'x':
                ptsonaxis = (tfracs, op1pts, op2pts)
                ptsoff1 = (tfracs, op1pts2, op2pts)
                ptsoff2 = (tfracs, op1pts, op2pts2)
            elif dirn == 'y':
                ptsonaxis = (op1pts, tfracs, op2pts)
                ptsoff1 = (op1pts2, tfracs, op2pts)
                ptsoff2 = (op1pts, tfracs, op2pts2)
            else:
                ptsonaxis = (op1pts, op2pts, tfracs)
                ptsoff1 = (op1pts2, op2pts, tfracs)
                ptsoff2 = (op1pts, op2pts2, tfracs)

            outstart += [
                N.ravel(N.column_stack(ptsonaxis)),
                N.ravel(N.column_stack(ptsonaxis))]
            outend += [
                N.ravel(N.column_stack(ptsoff1)),
                N.ravel(N.column_stack(ptsoff2))]

        # add labels for ticks and axis label
        if ticklabelsprop is not None:
            self.addLabels(
                cont, linecoords, ticklabelsprop, tfracs, tickvals,
                self.settings.label, self.settings.Label)

        # add ticks themselves
        if not tickprops.hide:
            startpts = threed.ValVector(N.concatenate(outstart))
            endpts = threed.ValVector(N.concatenate(outend))
            lineprop = tickprops.makeLineProp(painter)
            cont.addObject(threed.LineSegments(startpts, endpts, lineprop))

    def addGridLines(self, painter, cont, dirn, linecoords, gridprops, tickvals):
        """Add ticks for the vals and tick properties class given.
        linecoords: coordinates of start and end points of lines
        cont: container to add ticks
        dirn: 'x', 'y', 'z' for axis
        """

        if gridprops.hide:
            return

        tfracs = self.dataToLogicalCoords(tickvals, scaling=False)
        ones = N.ones(tfracs.shape)
        zeros = N.zeros(tfracs.shape)

        outstart = []
        outend = []

        # positions of grid lines for x axis
        pts1 = [
            (tfracs, zeros, zeros),
            (tfracs, zeros, zeros),
            (tfracs, ones, ones),
            (tfracs, ones, ones)
        ]
        pts2 = [
            (tfracs, zeros, ones),
            (tfracs, ones, zeros),
            (tfracs, zeros, ones),
            (tfracs, ones, zeros)
        ]
        # norm for each face, so we can draw back of cube only
        norms = [
            (0,  1,  0),
            (0,  0,  1),
            (0,  0, -1),
            (0, -1,  0)
        ]

        # swap coordinates for other axes
        if dirn == 'y':
            pts1  = [(c,a,b) for a,b,c in pts1]
            pts2  = [(c,a,b) for a,b,c in pts2]
            norms = [(c,a,b) for a,b,c in norms]
        elif dirn == 'z':
            pts1  = [(b,c,a) for a,b,c in pts1]
            pts2  = [(b,c,a) for a,b,c in pts2]
            norms = [(b,c,a) for a,b,c in norms]

        # add lines on each face
        lineprop = gridprops.makeLineProp(painter)
        for p1, p2, n in zip(pts1, pts2, norms):
            # container only shows face if norm points to observer
            face = threed.FacingContainer(threed.Vec3(*n))
            c1 = threed.ValVector(N.ravel(N.column_stack(p1)))
            c2 = threed.ValVector(N.ravel(N.column_stack(p2)))
            face.addObject(threed.LineSegments(c1, c2, lineprop))
            cont.addObject(face)

    def drawToObject(self, painter, painthelper):

        self.computePlottedRange()

        s = self.settings
        dirn = s.direction

        cont = threed.ObjectContainer()

        linecoords = self.addAxisLine(painter, cont, dirn)
        self.addAxisTicks(
            painter, cont, dirn, linecoords, s.MajorTicks, s.TickLabels,
            self.majortickscalc)
        self.addAxisTicks(
            painter, cont, dirn, linecoords, s.MinorTicks, None,
            self.minortickscalc)

        self.addGridLines(
            painter, cont, dirn, linecoords, s.GridLines,
            self.majortickscalc)
        self.addGridLines(
            painter, cont, dirn, linecoords, s.MinorGridLines,
            self.minortickscalc)

        cont.assignWidgetId(id(self))
        return cont

# allow the factory to instantiate an axis
document.thefactory.register(Axis3D)



# document.py
# A module to handle documents

#    Copyright (C) 2004 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""A class to represent Veusz documents, with dataset classes."""

import codecs
import os.path
import traceback
import datetime
from io import StringIO
from collections import defaultdict

try:
    import h5py
except ImportError:
    h5py = None

from .. import qtall as qt

from . import widgetfactory
from . import painthelper
from . import evaluate

from .. import datasets
from .. import utils
from .. import setting

def _(text, disambiguation=None, context="Document"):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

def getSuitableParent(widgettype, initialwidget):
    """Find the nearest relevant parent for the widgettype given."""

    # find the parent to add the child to, we go up the tree looking
    # for possible parents
    parent = initialwidget
    wc = widgetfactory.thefactory.getWidgetClass(widgettype)
    while parent is not None and not wc.willAllowParent(parent):
        parent = parent.parent
    return parent

class DocSuspend:
    """Handle document updates/suspensions."""
    def __init__(self, doc):
        self.doc = doc
    def __enter__(self):
        self.doc.suspendUpdates()
        return self
    def __exit__(self, type, value, traceback):
        self.doc.enableUpdates()

class Document(qt.QObject):
    """Document class for holding the graph data.
    """

    pluginsloaded = False

    # this is emitted when the document is modified
    signalModified = qt.pyqtSignal(int)
    # emited to log a message
    sigLog = qt.pyqtSignal(str)
    # emitted when document wiped
    sigWiped = qt.pyqtSignal()
    # to ask whether an import is allowed
    sigAllowedImports = qt.pyqtSignal()
    # security value set
    sigSecuritySet = qt.pyqtSignal(bool)

    # compatibility level
    maxcompatlevel = 1

    def __init__(self):
        """Initialise the document."""
        qt.QObject.__init__( self )

        if not Document.pluginsloaded:
            Document.loadPlugins()
            Document.pluginsloaded = True

        # change tracking of document as a whole
        self.changeset = 0            # increased when the document changes
        self.compatlevel = 0          # for non-backward compatible changes

        # map tags to dataset names
        self.datasettags = defaultdict(list)

        # if set, do not notify listeners of updates
        # wait under enableUpdates
        self.suspendupdates = []

        # default document locale
        self.locale = qt.QLocale()

        # evaluation context
        self.evaluate = evaluate.Evaluate(self)

        self.clearHistory()
        self.wipe()

    def wipe(self):
        """Wipe out any stored data."""
        self.data = {}

        self.basewidget = widgetfactory.thefactory.makeWidget(
            'document', None, self)

        self.setModified(False)
        self.filename = ""
        self.evaluate.wipe()
        self.sigWiped.emit()

    def clearHistory(self):
        """Clear any history."""
        self.historybatch = []
        self.historyundo = []
        self.historyredo = []

    def suspendUpdates(self):
        """Holds sending update messages.
        This speeds up modification of the document and prevents the document
        from being updated on the screen."""
        self.suspendupdates.append(self.changeset)

    def enableUpdates(self):
        """Reenables document updates."""
        changeset = self.suspendupdates.pop()
        if not self.suspendupdates and changeset != self.changeset:
            # bump this up as some watchers might ignore this otherwise
            self.changeset += 1
            self.setModified()

    def suspend(self):
        """Return context manager for suspending updates."""
        return DocSuspend(self)

    def setCompatLevel(self, level):
        """Update the compatiblity level.

        If level<0, then use latest version
        """
        if level < 0:
            level = self.maxcompatlevel
        elif level > self.maxcompatlevel:
            raise ValueError(
                "Compatibility level greater than supported by this Veusz version")
        self.compatlevel = level
        self.basewidget.makeStylesheet(compatlevel=self.compatlevel)

    def makeDefaultDoc(self, mode='graph', compatlevel=-1):
        """Add default widgets to create document.

        mode == 'graph', 'polar', 'ternary' or 'graph3d'
        """
        self.setCompatLevel(compatlevel)
        page = widgetfactory.thefactory.makeWidget(
            'page', self.basewidget, self)

        if mode == 'graph3d':
            scene = widgetfactory.thefactory.makeWidget('scene3d', page, self)
            widgetfactory.thefactory.makeWidget('graph3d', scene, self)
        else:
            assert mode in ('graph', 'polar', 'ternary')
            widgetfactory.thefactory.makeWidget(mode, page, self)
        self.setModified()
        self.setModified(False)
        self.changeset = 0
        self.evaluate.setSecurity(True)

    def log(self, message):
        """Log a message - this is emitted as a signal."""
        self.sigLog.emit(message)

    def applyOperation(self, operation, redoing=False):
        """Apply operation to the document.

        Operations represent atomic actions which can be done to the document
        and undone.

        Updates are suspended during the operation.

        If redoing is not True, the redo stack is cleared
        """

        with DocSuspend(self):
            retn = operation.do(self)
            self.changeset += 1

        if self.historybatch:
            # in batch mode, create an OperationMultiple for all changes
            self.historybatch[-1].addOperation(operation)
        else:
            # standard mode
            self.historyundo = self.historyundo[-9:] + [operation]

        if not redoing:
            self.historyredo = []

        return retn

    def batchHistory(self, batch):
        """Enable/disable batch history mode.

        In this mode further operations are added to the OperationMultiple specified,
        until batchHistory is called with None.

        The objects are pushed into a list and popped off

        This allows multiple operations to be batched up for simple undo.
        """
        if batch:
            self.historybatch.append(batch)
        else:
            self.historybatch.pop()

    def undoOperation(self):
        """Undo the previous operation."""

        operation = self.historyundo.pop()
        with DocSuspend(self):
            operation.undo(self)
            self.changeset += 1
        self.historyredo.append(operation)

    def canUndo(self):
        """Returns True if previous operation can be removed."""
        return len(self.historyundo) != 0

    def redoOperation(self):
        """Redo undone operations."""
        operation = self.historyredo.pop()
        return self.applyOperation(operation, redoing=True)

    def canRedo(self):
        """Returns True if previous operation can be redone."""
        return len(self.historyredo) != 0

    def isBlank(self):
        """Is the document unchanged?"""
        return self.changeset == 0

    def setData(self, name, dataset):
        """Set dataset in document."""
        self.data[name] = dataset
        dataset.document = self

        # update the change tracking
        self.setModified()

    def deleteData(self, name):
        """Remove a dataset"""
        del self.data[name]
        self.setModified()

    def modifiedData(self, dataset):
        """Notify dataset was modified"""
        assert dataset in self.data.values()
        self.setModified()

    def getLinkedFiles(self, filenames=None):
        """Get a list of LinkedFile objects used by the document.
        if filenames is a set, only get the objects with filenames given
        """
        links = set()
        for ds in self.data.values():
            if ds.linked and (
                    filenames is None or
                    ds.linked.filename in filenames):
                links.add(ds.linked)
        return list(links)

    def reloadLinkedDatasets(self, filenames=None):
        """Reload linked datasets from their files.
        If filenames is a set(), only reload from these filenames

        Returns a tuple of
        - List of datasets read
        - Dict of tuples containing dataset names and number of errors
        """

        links = self.getLinkedFiles(filenames=filenames)

        read = []
        errors = {}

        # load in the files, merging the vars read and errors
        if links:
            with self.suspend():
                for lf in links:
                    nread, nerrors = lf.reloadLinks(self)
                    read += nread
                    errors.update(nerrors)
                self.setModified()

        read.sort()
        return (read, errors)

    def datasetName(self, dataset):
        """Find name for given dataset, raising ValueError if missing."""
        for name, ds in self.data.items():
            if ds is dataset:
                return name
        raise ValueError("Cannot find dataset")

    def renameDataset(self, oldname, newname):
        """Rename the dataset."""
        d = self.data[oldname]
        del self.data[oldname]
        self.data[newname] = d

        self.setModified()

    def getData(self, name):
        """Get data with name"""
        return self.data[name]

    def setModified(self, ismodified=True):
        """Set the modified flag on the data, and inform views."""

        # useful for tracking back modifications
        # import traceback
        # traceback.print_stack()

        self.modified = ismodified
        self.changeset += 1

        if len(self.suspendupdates) == 0:
            self.signalModified.emit(ismodified)

    def isModified(self):
        """Return whether modified flag set."""
        return self.modified

    @classmethod
    def loadPlugins(kls, pluginlist=None):
        """Load plugins and catch exceptions."""
        if pluginlist is None:
            pluginlist = setting.settingdb.get('plugins', [])

        for plugin in pluginlist:
            try:
                with open(plugin) as f:
                    exec(f.read(), {})
            except Exception:
                err = _('Error loading plugin %s\n\n%s') % (
                    plugin, traceback.format_exc())
                raise RuntimeError(err)

    def paintTo(self, painthelper, page):
        """Paint page specified to the paint helper."""
        self.basewidget.draw(painthelper, page)

    def getNumberPages(self):
        """Return the number of pages in the document."""
        return len(self.basewidget.children)

    def getVisiblePages(self):
        """Return list of 0-indexed numbers of visible pages."""
        return [
            i for i, pg in enumerate(self.basewidget.children)
            if not pg.settings.hide
        ]

    def getPage(self, pagenumber):
        """Return widget for page."""
        return self.basewidget.children[pagenumber]

    def datasetTags(self):
        """Get list of all tags in datasets."""
        tags = set()
        for dataset in self.data.values():
            tags.update(dataset.tags)
        return sorted(tags)

    def _writeFileHeader(self, fileobj, type):
        """Write a header to a saved file of type."""

        fileobj.write('# Veusz %s (version %s)\n' % (type, utils.version()))
        fileobj.write(
            '# Saved at %s\n\n' %
            datetime.datetime.utcnow().isoformat())

    def saveDatasetTags(self, fileobj):
        """Write dataset tags to output file"""

        # get a list of all tags and which datasets have them
        bytag = defaultdict(list)
        for name, dataset in sorted(self.data.items()):
            for t in dataset.tags:
                bytag[t].append(name)

        # write out tags
        for tag, val in sorted(bytag.items()):
            fileobj.write(
                'TagDatasets(%s, %s)\n' % (utils.rrepr(tag), utils.rrepr(val)))

    def saveToFile(self, fileobj):
        """Save the text representing a document to a file.

        The ordering can be important, as some things override
        previous steps:

         - Tagging doesn't work if the dataset isn't
           already defined.
         - Loading from files may bring in new datasets which
           override defined datasets, so save links first
        """

        self._writeFileHeader(fileobj, 'saved document')

        # write compatibility level
        fileobj.write('SetCompatLevel(%s)\n' % self.compatlevel)

        # add file directory to import path if we know it
        reldirname = None
        if getattr(fileobj, 'name', False):
            reldirname = os.path.dirname( os.path.abspath(fileobj.name) )
            if setting.settingdb['docfile_addimportpaths']:
                fileobj.write('AddImportPath(%s)\n' % utils.rrepr(reldirname))

        # save those datasets which are linked
        # we do this first in case the datasets are overridden below
        savedlinks = {}
        for name, dataset in sorted(self.data.items()):
            dataset.saveLinksToSavedDoc(fileobj, savedlinks,
                                        relpath=reldirname)

        # save the remaining datasets
        for name, dataset in sorted(self.data.items()):
            dataset.saveToFile(fileobj, name)

        # save tags of datasets
        self.saveDatasetTags(fileobj)

        # add custom definitions
        self.evaluate.saveCustomDefinitions(fileobj)

        # save the actual tree structure
        fileobj.write(self.basewidget.getSaveText())

        self.setModified(False)

    def saveToHDF5File(self, fileobj):
        """Save to HDF5 (h5py) output file given."""

        # groups in output hdf5
        vszgrp = fileobj.create_group('Veusz')
        vszgrp.attrs['vsz_version'] = utils.version()
        vszgrp.attrs['vsz_saved_at'] = datetime.datetime.utcnow().isoformat()
        vszgrp.attrs['vsz_format'] = 1  # version number (currently unused)
        datagrp = vszgrp.create_group('Data')
        docgrp = vszgrp.create_group('Document')

        textstream = StringIO()

        self._writeFileHeader(textstream, 'saved document')

        # add file directory to import path if we know it
        reldirname = None
        if getattr(fileobj, 'filename', False):
            reldirname = os.path.dirname( os.path.abspath(fileobj.filename) )
            textstream.write('AddImportPath(%s)\n' % utils.rrepr(reldirname))

        # add custom definitions
        self.evaluate.saveCustomDefinitions(textstream)

        # save those datasets which are linked
        # we do this first in case the datasets are overridden below
        savedlinks = {}
        for name, dataset in sorted(self.data.items()):
            dataset.saveLinksToSavedDoc(
                textstream, savedlinks, relpath=reldirname)

        # save the remaining datasets
        for name, dataset in sorted(self.data.items()):
            dataset.saveToFile(textstream, name, mode='hdf5', hdfgroup=datagrp)

        # handle tagging
        # get a list of all tags and which datasets have them
        bytag = defaultdict(list)
        for name, dataset in sorted(self.data.items()):
            for t in dataset.tags:
                bytag[t].append(name)

        # write out tags as datasets
        tagsgrp = docgrp.create_group('Tags')
        for tag, dsnames in sorted(bytag.items()):
            tagsgrp[tag] = [v.encode('utf-8') for v in sorted(dsnames)]

        # save the actual tree structure
        textstream.write(self.basewidget.getSaveText())

        # create single dataset contains document
        docgrp['document'] = [ textstream.getvalue().encode('utf-8') ]

        self.setModified(False)

    def save(self, filename, mode='vsz'):
        """Save to output file.

        mode is 'vsz' or 'hdf5'
        """
        if mode == 'vsz':
            with codecs.open(filename, 'w', 'utf-8') as f:
                self.saveToFile(f)
        elif mode == 'hdf5':
            if h5py is None:
                raise RuntimeError('Missing h5py module')
            with h5py.File(filename, 'w') as f:
                self.saveToHDF5File(f)
        else:
            raise RuntimeError('Invalid save mode')

        self.filename = filename

    def load(self, filename, mode='vsz',
             callbackunsafe=None,
             callbackimporterror=None):
        """Load document from file.

        mode is 'vsz' or 'hdf5'
        """
        from . import loader
        loader.loadDocument(
            self, filename, mode=mode,
            callbackunsafe=callbackunsafe,
            callbackimporterror=callbackimporterror)

    def exportStyleSheet(self, fileobj):
        """Export the StyleSheet to a file."""

        self._writeFileHeader(fileobj, 'exported stylesheet')
        stylesheet = self.basewidget.settings.StyleSheet

        fileobj.write( stylesheet.saveText(True, rootname='') )

    def _pagedocsize(self, widget, dpi, scaling, integer):
        """Helper for page or doc size."""
        if dpi is None:
            p = qt.QPixmap(1, 1)
            dpi = (p.logicalDpiX(), p.logicalDpiY())
        helper = painthelper.PaintHelper(self, (1,1), dpi=dpi)
        w = widget.settings.get('width').convert(helper) * scaling
        h = widget.settings.get('height').convert(helper) * scaling
        if integer:
            return int(w), int(h)
        else:
            return w, h

    def pageSize(self, pagenum, dpi=None, scaling=1., integer=True):
        """Get the size of a particular page in pixels.

        If dpi is None, use the default Qt screen dpi
        Use dpi if given."""

        page = self.basewidget.getPage(pagenum)
        if page is None:
            return self.docSize(dpi=dpi, scaling=scaling, integer=integer)
        return self._pagedocsize(
            page, dpi=dpi, scaling=scaling, integer=integer)

    def docSize(self, dpi=None, scaling=1., integer=True):
        """Get size for document."""
        return self._pagedocsize(
            self.basewidget,
            dpi=dpi, scaling=scaling, integer=integer)

    def resolvePath(self, fromobj, path):
        """Resolve item relative to fromobj.

        If fromobj is None, then an absolute path is assumed.

        Returns a widget, setting or settings as appropriate.
        """

        # where to search from
        obj = self.basewidget if (path[:1]=='/' or fromobj is None) else fromobj

        # iterate over path parts
        for p in path.split('/'):
            if p == '..':
                p = obj.parent
                if p is None:
                    raise ValueError("Base graph has no parent")
                obj = p
            elif p == '.' or len(p) == 0:
                pass
            elif obj.iswidget:
                if p in obj.settings:
                    obj = obj.settings.get(p)
                else:
                    widget = obj.getChild(p)
                    if widget is None:
                        raise ValueError("Widget has no child %s" % p)
                    obj = widget
            elif obj.issettings:
                try:
                    obj = obj.get(p)
                except KeyError:
                    raise ValueError("Settings has no child %s" % p)
            else:
                raise ValueError("Item has no children")

        return obj

    def resolveWidgetPath(self, fromobj, path):
        """Resolve path to Widget.

        If fromobj is None, then is resolved to root widget.
        Raises ValueError if path is invalid or not to widget.
        """
        obj = self.resolvePath(fromobj, path)
        if not obj.iswidget:
            raise ValueError("Not path to widget")
        return obj

    def resolveSettingPath(self, fromobj, path):
        """Resolve path to Setting.

        If fromobj is None, then is resolved to root widget.
        Raises ValueError if path is invalid or not to Setting.
        """
        obj = self.resolvePath(fromobj, path)
        if not obj.issetting:
            raise ValueError("Not path to setting")
        return obj

    def walkNodes(self, tocall, root=None,
                  nodetypes=('widget', 'setting', 'settings'),
                  _path=None):
        """Walk the widget/settings/setting nodes in the document.
        For each one call tocall(path, node).
        nodetypes is tuple of possible node types
        """

        if root is None:
            root = self.basewidget
        if _path is None:
            _path = root.path

        if root.nodetype in nodetypes:
            tocall(_path, root)

        if root.nodetype == 'widget':
            # get rid of // at start of path
            if _path == '/':
                _path = ''

            # do the widget's children
            for w in root.children:
                self.walkNodes(
                    tocall, root=w, nodetypes=nodetypes,
                    _path=_path+'/'+w.name)
            # then do the widget's settings
            self.walkNodes(
                tocall, root=root.settings,
                nodetypes=nodetypes, _path=_path)
        elif root.nodetype == 'settings':
            # do the settings of the settings
            for name, s in sorted(root.setdict.items()):
                self.walkNodes(
                    tocall, root=s, nodetypes=nodetypes,
                    _path=_path+'/'+s.name)
        # elif root.nodetype == 'setting': pass

    def formatValsWithDatatypeToText(self, vals, displaydatatype):
        """Given a set of values, datatype, return a list of strings
        corresponding to these data."""

        if displaydatatype == 'text':
            return vals
        elif displaydatatype == 'numeric':
            return [
                utils.formatNumber(val, '%Vg', locale=self.locale)
                for val in vals ]
        elif displaydatatype == 'date':
            return [ utils.dateFloatToString(val) for val in vals ]
        else:
            raise RuntimeError('Invalid data type')



#    Copyright (C) 2013 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

from .. import qtall as qt
from .. import utils
from .. import document
from . import simpleread
from . import base

def _(text, disambiguation=None, context="Import_2D"):
    return qt.QCoreApplication.translate(context, text, disambiguation)

class ImportParams2D(base.ImportParamsBase):
    """2D import parameters.

    additional parameters:
     datastr: text to read from instead of file
     xrange: tuple with range of x data coordinates
     yrange: tuple with range of y data coordinates
     invertrows: invert rows when reading
     invertcols: invert columns when reading
     transpose: swap rows and columns
     gridatedge: use left column and top row for pixel centres
     mode: text or csv
     csvdelimiter/csvtextdelimiter: csv text delimiters
     csvlocale: locale when importing csv
    """

    defaults = {
        'datasetnames': None,
        'datastr': None,
        'xrange': None,
        'yrange': None,
        'invertrows': False,
        'invertcols': False,
        'transpose': False,
        'gridatedge': False,
        'mode': 'text',
        'csvdelimiter': ',',
        'csvtextdelimiter': '"',
        'csvlocale': 'en_US',
    }
    defaults.update(base.ImportParamsBase.defaults)

class LinkedFile2D(base.LinkedFileBase):
    """Class representing a file linked to a 2d dataset."""

    def createOperation(self):
        """Return operation to recreate self."""
        return OperationDataImport2D

    def saveToFile(self, fileobj, relpath=None):
        """Save the link to the document file."""
        self._saveHelper(
            fileobj,
            'ImportFile2D',
            ('filename', 'datasetnames'),
            relpath=relpath)

class OperationDataImport2D(base.OperationDataImportBase):
    """Import a 2D matrix from a file."""

    descr = _('import 2D data')

    def doImport(self):
        """Import data."""

        p = self.params

        # get stream
        if p.mode == 'csv':
            stream = simpleread.CSVStream(
                p.filename, p.csvdelimiter, p.csvtextdelimiter,
                p.csvlocale, p.encoding)
        elif p.filename is not None:
            stream = simpleread.FileStream(
                utils.openEncoding(p.filename, p.encoding) )
        elif p.datastr is not None:
            stream = simpleread.StringStream(p.datastr)
        else:
            raise RuntimeError("Invalid combination of parameters")

        # linked file
        LF = None
        if p.linked:
            assert p.filename
            LF = LinkedFile2D(p)

        for name in p.datasetnames:
            sr = simpleread.SimpleRead2D(name, p)
            sr.readData(stream)
            sr.setOutput(self.outdatasets, linkedfile=LF)

def ImportFile2D(comm, filename, datasetnames, xrange=None, yrange=None,
                 invertrows=None, invertcols=None, transpose=None,
                 gridatedge=None,
                 mode='text', csvdelimiter=',', csvtextdelimiter='"',
                 csvlocale='en_US',
                 prefix="", suffix="", encoding='utf_8',
                 renames=None,
                 linked=False):
    """Import two-dimensional data from a file.
    filename is the name of the file to read
    datasetnames is a list of datasets to read from the file, or a single
    dataset name

    xrange is a tuple containing the range of data in x coordinates
    yrange is a tuple containing the range of data in y coordinates
    if invertrows=True, then rows are inverted when read
    if invertcols=True, then cols are inverted when read
    if transpose=True, then rows and columns are swapped
    if gridatedge=True, use top row and left column for pixel positions

    mode is either 'text' or 'csv'
    csvdelimiter is the csv delimiter for csv
    csvtextdelimiter is the csv text delimiter for csv
    csvlocale is locale to use when reading csv data

    prefix and suffix are prepended and appended to dataset names

    encoding is encoding character set

    renames is an optional dict mapping old to new dataset names, to
    be renamed after importing

    if linked=True then the dataset is linked to the file

    Returns: list of imported datasets
    """

    # look up filename on path
    realfilename = comm.findFileOnImportPath(filename)

    if isinstance(datasetnames, str):
        datasetnames = [datasetnames]

    params = ImportParams2D(
        datasetnames=datasetnames,
        filename=realfilename, xrange=xrange,
        yrange=yrange, invertrows=invertrows,
        invertcols=invertcols, transpose=transpose,
        gridatedge=gridatedge,
        mode=mode,
        csvdelimiter=csvdelimiter,
        csvtextdelimiter=csvtextdelimiter,
        csvlocale=csvlocale,
        prefix=prefix, suffix=suffix,
        renames=renames,
        linked=linked)
    op = OperationDataImport2D(params)
    comm.document.applyOperation(op)

    if comm.verbose:
        print("Imported datasets %s" % ', '.join(op.outnames))
    return op.outnames

def ImportString2D(comm, datasetnames, dstring, xrange=None, yrange=None,
                   invertrows=None, invertcols=None, transpose=None):
    """Read two dimensional data from the string specified.
    datasetnames is a list of datasets to read from the string or a single
    dataset name

    xrange is a tuple containing the range of data in x coordinates
    yrange is a tuple containing the range of data in y coordinates
    if invertrows=True, then rows are inverted when read
    if invertcols=True, then cols are inverted when read
    if transpose=True, then rows and columns are swapped

    Returns: list of imported datasets
    """

    if isinstance(datasetnames, str):
        datasetnames = [datasetnames]

    params = ImportParams2D(
        datasetnames=datasetnames,
        datastr=dstring, xrange=xrange,
        yrange=yrange, invertrows=invertrows,
        invertcols=invertcols, transpose=transpose)
    op = OperationDataImport2D(params)
    comm.document.applyOperation(op)

    if comm.verbose:
        print("Imported datasets %s" % ', '.join(op.outnames))
    return op.outnames

document.registerImportCommand('ImportFile2D', ImportFile2D)
document.registerImportCommand(
    'ImportString2D', ImportString2D, filenamearg=-1)



#    Copyright (C) 2013 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

'''An axis which can be broken in places.'''

import bisect

import numpy as N

from .. import qtall as qt
from .. import setting
from .. import document
from .. import utils

from . import axis
from . import controlgraph

def _(text, disambiguation=None, context='BrokenAxis'):
    '''Translate text.'''
    return qt.QCoreApplication.translate(context, text, disambiguation)

class AxisBroken(axis.Axis):
    '''An axis widget which can have gaps in it.'''

    typename = 'axis-broken'
    description = 'Axis with breaks in it'

    def __init__(self, parent, name=None):
        """Initialise axis."""
        axis.Axis.__init__(self, parent, name=name)
        self.rangeswitch = None
        self.breakchangeset = -1

    @classmethod
    def addSettings(klass, s):
        '''Construct list of settings.'''
        axis.Axis.addSettings(s)

        s.add( setting.FloatList(
            'breakPoints',
            [],
            descr = _('Pairs of values to start and stop breaks'),
            usertext = _('Break pairs')), 4 )
        s.add( setting.FloatList(
            'breakPosns',
            [],
            descr = _('Positions (fractions) along axis where to break'),
            usertext = _('Break positions'),
            formatting=True) )

    def switchBreak(self, num, posn, otherposition=None):
        """Switch to break given (or None to disable)."""
        self.rangeswitch = num
        if num is None:
            self.plottedrange = self.orig_plottedrange
        else:
            self.plottedrange = [self.breakvstarts[num], self.breakvstops[num]]
        self.updateAxisLocation(posn, otherposition=otherposition)

    def plotterToGraphCoords(self, bounds, vals):
        """Convert values in plotter coordinates to data values.  This
        needs to know about whether we've not switched between the
        breaks.

        Note that this implementation is very slow! Hopefully it won't
        be called often.
        """

        if self.rangeswitch is not None:
            return axis.Axis.plotterToGraphCoords(self, bounds, vals)

        # support single int/float values
        try:
            iter(vals)
            issingle = False
        except TypeError:
            vals = N.array([vals])
            issingle = True

        # scaled to be fractional coordinates in bounds
        if self.settings.direction == 'horizontal':
            svals = (vals - bounds[0]) / (bounds[2] - bounds[0])
        else:
            svals = (vals - bounds[3]) / (bounds[1] - bounds[3])

        # first work out which break region the values are in
        out = []
        for sval, val in zip(svals, vals):
            # find index for appropriated scaled starting value
            breaki = bisect.bisect_left(self.posstarts, sval) - 1

            if ( breaki >= 0 and breaki < self.breakvnum and
                 sval <= self.posstops[breaki] ):
                self.switchBreak(breaki, bounds)
                coord = axis.Axis.plotterToGraphCoords(
                    self, bounds, N.array([val]))[0]
            else:
                coord = N.nan
            out.append(coord)
        self.switchBreak(None, bounds)

        if issingle:
            return out[0]
        else:
            return N.array(out)

    def _graphToPlotter(self, vals):
        """Convert graph values to plotter coords.
        This could be slow if no range selected
        """

        if self.rangeswitch is not None:
            return axis.Axis._graphToPlotter(self, vals)

        # support single int/float values
        try:
            iter(vals)
            issingle = False
        except TypeError:
            vals = N.array([vals])
            issingle = True

        out = []
        for val in vals:
            breaki = bisect.bisect_left(self.breakvstarts, val) - 1
            if breaki >= 0 and breaki < self.breakvnum:
                if val > self.breakvstops[breaki] and breaki < self.breakvnum-1:
                    # in gap, so use half-value
                    coord = 0.5*(self.posstops[breaki]+self.posstarts[breaki+1])

                    b = self.currentbounds
                    if self.settings.direction == 'horizontal':
                        coord = coord*(b[2] - b[0]) + b[0]
                    else:
                        coord = coord*(b[3] - b[1]) + b[1]

                else:
                    # lookup value
                    self.switchBreak(breaki, self.currentbounds)
                    coord = axis.Axis._graphToPlotter(self, N.array([val]))[0]
            else:
                coord = N.nan
            out.append(coord)
        self.switchBreak(None, self.currentbounds)

        if issingle:
            return out[0]
        else:
            return N.array(out)

    def updateAxisLocation(self, bounds, otherposition=None):
        """Recalculate broken axis positions."""

        s = self.settings

        if self.document.changeset != self.breakchangeset:
            self.breakchangeset = self.document.changeset

            # actually start and stop values on axis
            num = len(s.breakPoints) // 2
            posns = list(s.breakPosns)
            posns.sort()

            # add on more break positions if not specified
            if len(posns) < num:
                start = 0.
                if len(posns) != 0:
                    start = posns[-1]
                posns = posns + list(
                    N.arange(1,num-len(posns)+1) *
                    ( (1.-start) / (num-len(posns)+1) + start ))

            # fractional difference between starts and stops
            breakgap = 0.05

            # collate fractional positions for starting and stopping
            starts = [0.]
            stops = []
            for pos in posns:
                stops.append( pos - breakgap/2. )
                starts.append( pos + breakgap/2. )
            stops.append(1.)

            # scale according to allowable range
            d = s.upperPosition - s.lowerPosition
            self.posstarts = N.array(starts)*d + s.lowerPosition
            self.posstops = N.array(stops)*d + s.lowerPosition

        # pass lower and upper ranges if a particular range is chosen
        if self.rangeswitch is None:
            lowerupper = None
        else:
            lowerupper = ( self.posstarts[self.rangeswitch],
                           self.posstops[self.rangeswitch] )

        axis.Axis.updateAxisLocation(
            self, bounds,
            otherposition=otherposition,
            lowerupperposition=lowerupper)

    def computePlottedRange(self):
        """Given range of data, recompute stops and start values of
        breaks."""

        axis.Axis.computePlottedRange(self)

        r = self.orig_plottedrange = self.plottedrange
        points = list(self.settings.breakPoints)
        points.sort()
        if r[1] < r[0]:
            points.reverse()

        # filter to range
        newpoints = []
        for i in range(0, len(points)//2 * 2, 2):
            if points[i] >= min(r) and points[i+1] <= max(r):
                newpoints += [points[i], points[i+1]]

        self.breakvnum = num = len(newpoints)//2 + 1
        self.breakvlist = [self.plottedrange[0]] + newpoints + [
            self.plottedrange[1]]

        # axis values for starting and stopping
        self.breakvstarts = [ self.breakvlist[i*2] for i in range(num) ]
        self.breakvstops = [ self.breakvlist[i*2+1] for i in range(num) ]

        # compute ticks for each range
        self.minorticklist = []
        self.majorticklist = []
        for i in range(self.breakvnum):
            self.plottedrange = [self.breakvstarts[i], self.breakvstops[i]]
            reverse = self.plottedrange[0] > self.plottedrange[1]
            if reverse:
                self.plottedrange.reverse()
            self.computeTicks(allowauto=False)
            if reverse:
                self.plottedrange.reverse()
            self.minorticklist.append(self.minortickscalc)
            self.majorticklist.append(self.majortickscalc)

        self.plottedrange = self.orig_plottedrange

    def _drawAutoMirrorTicks(self, posn, painter):
        """Mirror axis to opposite side of graph if there isn't an
        axis there already."""

        # swap axis to other side
        s = self.settings
        if s.otherPosition < 0.5:
            otheredge = 1.
        else:
            otheredge = 0.

        # temporarily change position of axis to other side for drawing
        self.updateAxisLocation(posn, otherposition=otheredge)
        if not s.Line.hide:
            self._drawAxisLine(painter, posn)

        for i in range(self.breakvnum):
            self.switchBreak(i, posn, otherposition=otheredge)

            # plot coordinates of ticks
            coordticks = self._graphToPlotter(self.majorticklist[i])
            coordminorticks = self._graphToPlotter(self.minorticklist[i])

            if not s.MinorTicks.hide:
                self._drawMinorTicks(painter, coordminorticks)
            if not s.MajorTicks.hide:
                self._drawMajorTicks(painter, coordticks)

        self.switchBreak(None, posn)

    def _drawAxisLine(self, painter, posn):
        """Draw the line of the axis, indicating broken positions.
        We currently use a triangle to mark the broken position
        """

        # these are x and y, or y and x coordinates
        p1 = [self.posstarts[0]]
        p2 = [0.]

        # mirror shape using this setting
        markdirn = -1
        if self.coordReflected:
            markdirn = -markdirn

        # add shape for each break
        for start, stop in zip( self.posstarts[1:], self.posstops[:-1] ):
            p1 += [stop, (start+stop)*0.5, start]
            p2 += [0, markdirn*(start-stop)*0.5, 0]

        # end point
        p1.append(self.posstops[-1])
        p2.append(0.)

        # scale points by length of axis and add correct origin
        s = self.settings

        if s.direction == 'vertical':
            delta = posn[1]-posn[3]
            minv = posn[3]
        else:
            delta = posn[2]-posn[0]
            minv = posn[0]

        p1 = N.array(p1)*delta + minv
        p2 = N.array(p2)*delta + self.coordPerp

        if s.direction == 'vertical':
            p1, p2 = p2, p1

        # convert to polygon and draw
        poly = qt.QPolygonF()
        utils.addNumpyToPolygonF(poly, p1, p2)

        pen = s.get('Line').makeQPen(painter)
        pen.setCapStyle(qt.Qt.FlatCap)
        painter.setPen(pen)
        painter.drawPolyline(poly)

    def drawGrid(self, parentposn, phelper, outerbounds=None,
                 ontop=False):
        """Code to draw gridlines.

        This is separate from the main draw routine because the grid
        should be behind/infront the data points.
        """

        s = self.settings
        if ( s.hide or (s.MinorGridLines.hide and s.GridLines.hide) or
             s.GridLines.onTop != bool(ontop) ):
            return

        # draw grid on a different layer, depending on whether on top or not
        layer = (-2, -1)[bool(ontop)]
        painter = phelper.painter(self, parentposn, layer=layer)
        self.updateAxisLocation(parentposn)

        with painter:
            painter.save()
            painter.setClipRect( qt.QRectF(
                qt.QPointF(parentposn[0], parentposn[1]),
                qt.QPointF(parentposn[2], parentposn[3]) ) )

            for i in range(self.breakvnum):
                self.switchBreak(i, parentposn)
                if not s.MinorGridLines.hide:
                    coordminorticks = self._graphToPlotter(self.minorticklist[i])
                    self._drawGridLines(
                        'MinorGridLines', painter, coordminorticks,
                        parentposn)
                if not s.GridLines.hide:
                    coordticks = self._graphToPlotter(self.majorticklist[i])
                    self._drawGridLines(
                        'GridLines', painter, coordticks,
                        parentposn)

            self.switchBreak(None, parentposn)
            painter.restore()

    def _axisDraw(self, posn, parentposn, outerbounds, painter, phelper):
        """Main drawing routine of axis."""

        s = self.settings

        # multiplication factor if reflection on the axis is requested
        sign = 1
        if s.direction == 'vertical':
            sign *= -1
        if self.coordReflected:
            sign *= -1

        # keep track of distance from axis
        # text to output
        texttorender = []

        # plot the line along the axis
        if not s.Line.hide:
            self._drawAxisLine(painter, posn)

        max_delta = 0
        suppresstext = False
        for i in range(self.breakvnum):
            self.switchBreak(i, posn)

            if self.plottedrange[0]==self.plottedrange[1]:
                continue

            # plot coordinates of ticks
            coordticks = self._graphToPlotter(self.majorticklist[i])
            coordminorticks = self._graphToPlotter(self.minorticklist[i])

            self._delta_axis = 0

            # plot minor ticks
            if not s.MinorTicks.hide:
                self._drawMinorTicks(painter, coordminorticks)

            # plot major ticks
            if not s.MajorTicks.hide:
                self._drawMajorTicks(painter, coordticks)

            # plot tick labels
            suppresstext = self._suppressText(painter, parentposn, outerbounds)
            if not s.TickLabels.hide and not suppresstext:
                self._drawTickLabels(
                    phelper, painter, coordticks, sign,
                    outerbounds, self.majorticklist[i],
                    texttorender)

            # this is the maximum delta of any of the breaks
            max_delta = max(max_delta, self._delta_axis)

        self.switchBreak(None, posn)
        self._delta_axis = max_delta

        # draw an axis label
        if not s.Label.hide and not suppresstext:
            self._drawAxisLabel(painter, sign, outerbounds, texttorender)

        self._drawTextWithoutOverlap(painter, texttorender)

        # make control item for axis
        phelper.setControlGraph(self, [ controlgraph.ControlAxisLine(
            self, phelper, self.settings.direction, self.coordParr1,
            self.coordParr2, self.coordPerp, posn) ])

# allow the factory to instantiate the widget
document.thefactory.register(AxisBroken)



#!/usr/bin/env python3

#    Copyright (C) 2008 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""
Veusz distutils setup script
see the file INSTALL for details on how to install Veusz
"""

import sys
import glob
import os.path
import numpy

import setuptools
from setuptools import setup, Extension
from setuptools.command.install import install as orig_install
from distutils.command.install_data import install_data
import pyqtdistutils

if sys.version_info[0] < 3:
    raise RuntimeError('Veusz only supports Python 3')

# get version
with open('VERSION') as verf:
    version = verf.read().strip()

class install(orig_install):
    user_options = orig_install.user_options + [
        # tell veusz where to install its data files
        (
            'veusz-resource-dir=', None,
            'override veusz resource directory location'
        ),
        (
            'disable-install-examples', None,
            'do not install examples files'
        ),
    ]
    boolean_options = orig_install.boolean_options + [
        'disable-install-examples',
    ]

    def initialize_options(self):
        orig_install.initialize_options(self)
        self.veusz_resource_dir = None
        self.disable_install_examples = False

# Pete Shinner's distutils data file fix... from distutils-sig
#  data installer with improved intelligence over distutils
#  data files are copied into the project directory instead
#  of willy-nilly
class smart_install_data(install_data):
    def run(self):
        install_cmd = self.get_finalized_command('install')
        if install_cmd.veusz_resource_dir:
            # override location with veusz-resource-dir option
            self.install_dir = install_cmd.veusz_resource_dir
        else:
            # change self.install_dir to the library dir + veusz by default
            self.install_dir = os.path.join(install_cmd.install_lib, 'veusz')

        # disable examples install if requested
        if install_cmd.disable_install_examples:
            self.data_files = [
                f for f in self.data_files if f[0][-8:] != 'examples'
            ]

        return install_data.run(self)

descr = '''Veusz is a 2D and 3D scientific plotting package, designed to
create publication-ready PDF and SVG output. It features GUI,
command-line, and scripting interfaces. Graphs are constructed from
"widgets", allowing complex layouts to be designed. Veusz supports
plotting functions, data with errors, keys, labels, stacked plots,
multiple plots, and fitting data.'''

def findData(dirname, extns):
    """Return tuple for directory name and list of file extensions for data."""
    files = []
    for extn in extns:
        files += glob.glob(os.path.join(dirname, '*.'+extn))
    files.sort()
    return (dirname, files)

setup(
    name = 'veusz',
    version = version,
    description = 'A scientific plotting package',
    long_description = descr,
    author = 'Jeremy Sanders',
    author_email = 'jeremy@jeremysanders.net',
    url = 'https://veusz.github.io/',
    license = 'GPLv2+',
    classifiers = [
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Development Status :: 5 - Production/Stable',
        'Environment :: X11 Applications :: Qt',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: '
        'GNU General Public License v2 or later (GPLv2+)',
        'Topic :: Scientific/Engineering :: Visualization'
    ],
    python_requires='>=3.3',

    data_files = [
        ('', ['VERSION', 'AUTHORS', 'ChangeLog', 'COPYING']),
        findData('ui', ('ui',)),
        findData('icons', ('png', 'svg')),
        findData('examples', ('vsz', 'py', 'csv', 'dat')),
    ],
    packages = [
        'veusz',
        'veusz.dataimport',
        'veusz.datasets',
        'veusz.dialogs',
        'veusz.document',
        'veusz.helpers',
        'veusz.plugins',
        'veusz.qtwidgets',
        'veusz.setting',
        'veusz.utils',
        'veusz.widgets',
        'veusz.windows',
    ],

    ext_modules = [
        # threed support
        Extension(
            'veusz.helpers.threed',
            [
                'veusz/helpers/src/threed/camera.cpp',
                'veusz/helpers/src/threed/mmaths.cpp',
                'veusz/helpers/src/threed/objects.cpp',
                'veusz/helpers/src/threed/scene.cpp',
                'veusz/helpers/src/threed/fragment.cpp',
                'veusz/helpers/src/threed/numpy_helpers.cpp',
                'veusz/helpers/src/threed/clipcontainer.cpp',
                'veusz/helpers/src/threed/bsp.cpp',
                'veusz/helpers/src/threed/twod.cpp',
                'veusz/helpers/src/threed/threed.sip'
            ],
            language="c++",
            include_dirs=[
                'veusz/helpers/src/threed', numpy.get_include()
            ],
        ),

        # mathml widget
        Extension(
            'veusz.helpers.qtmml',
            [
                'veusz/helpers/src/qtmml/qtmmlwidget.cpp',
                'veusz/helpers/src/qtmml/qtmml.sip'
            ],
            language="c++",
            include_dirs=['veusz/helpers/src/qtmml'],
        ),

        # device to record paint commands
        Extension(
            'veusz.helpers.recordpaint',
            [
                'veusz/helpers/src/recordpaint/recordpaintdevice.cpp',
                'veusz/helpers/src/recordpaint/recordpaintengine.cpp',
                'veusz/helpers/src/recordpaint/recordpaint.sip'
            ],
            language="c++",
            include_dirs=['veusz/helpers/src/recordpaint'],
        ),

        # contour plotting library
        Extension(
            'veusz.helpers._nc_cntr',
            [
                'veusz/helpers/src/nc_cntr/_nc_cntr.c'
            ],
            include_dirs=[numpy.get_include()]
        ),

        # qt helper module
        Extension(
            'veusz.helpers.qtloops',
            [
                'veusz/helpers/src/qtloops/qtloops.cpp',
                'veusz/helpers/src/qtloops/qtloops_helpers.cpp',
                'veusz/helpers/src/qtloops/polygonclip.cpp',
                'veusz/helpers/src/qtloops/polylineclip.cpp',
                'veusz/helpers/src/qtloops/beziers.cpp',
                'veusz/helpers/src/qtloops/beziers_qtwrap.cpp',
                'veusz/helpers/src/qtloops/numpyfuncs.cpp',
                'veusz/helpers/src/qtloops/qtloops.sip'],
            language="c++",
            include_dirs=[
                'veusz/helpers/src/qtloops',
                numpy.get_include()
            ],
        ),
    ],

    # new command options
    cmdclass = {
        'build_ext': pyqtdistutils.build_ext,
        'install_data': smart_install_data,
        'install': install
    },

    # requires these modules to install
    install_requires = [
        'numpy', 'PyQt5'
    ],

    # optional requirements
    extras_require = {
        "optional": [
            'astropy', 'pyemf3', 'sampy', 'iminuit', 'h5py'
        ]
    },

    # GUI entry points
    entry_points = {
        'gui_scripts' : [
            'veusz = veusz.veusz_main:run',
        ]
    },
)



#    Copyright (C) 2012 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
###############################################################################

"""Paint fills with extended brush class.

Paints solid, hatching and various qt brushes
"""

import math

import numpy as N

from .. import qtall as qt
from ..helpers.qtloops import plotLinesToPainter, polygonClip

def dumppath(p):
    i =0
    while i < p.elementCount():
        e = p.elementAt(i)
        if e.isLineTo():
            print(" l(%i,%i)" %( e.x, e.y), end=' ')
            i += 1
        elif e.isMoveTo():
            print(" m(%i,%i)" %(e.x, e.y), end=' ')
            i += 1
        else:
            print(" c(%i,%i,%i,%i,%i,%i)" %(e.x, e.y, p.elementAt(i+1).x, p.elementAt(i+1).y, p.elementAt(i+2).x, p.elementAt(i+2).y), end=' ')
            i += 3
    print()

def _hatcher(painter, pen, painterpath, spacing, hatchlist):
    """Draw hatching on painter path given."""

    painter.save()
    painter.setPen(pen)

    # debugging
    # dumppath(painterpath)

    painter.setClipPath(painterpath, qt.Qt.IntersectClip)

    # this is the bounding box of the path
    bb = painter.clipPath().boundingRect()

    for params in hatchlist:
        # scale values
        offsetx, offsety, deltax, deltay = [x*spacing for x in params]

        # compute max number of steps
        numsteps = 0
        if deltax != 0:
            numsteps += int(bb.width() / abs(deltax)) + 4
        if deltay != 0:
            numsteps += int(bb.height() / abs(deltay)) + 4

        # calculate repeat position in x and y
        # we start the positions of lines at multiples of these distances
        # to ensure the pattern doesn't shift if the area does

        # distances between lines
        mag = math.sqrt(deltax**2 + deltay**2)
        # angle to x of perpendicular lines
        theta = math.atan2(deltay, deltax)
        s, c = math.sin(theta), math.cos(theta)
        intervalx = intervaly = 1.
        if abs(c) > 1e-4: intervalx = abs(mag / c)
        if abs(s) > 1e-4: intervaly = abs(mag / s)

        startx = int(bb.left() / intervalx) * intervalx + offsetx
        starty = int(bb.top()  / intervaly) * intervaly + offsety

        # normalise line vector
        linedx, linedy = -deltay/mag, deltax/mag
        # scale of lines from start position
        scale = max( bb.width(), bb.height() ) * 4

        # construct points along lines
        # this is scales to ensure the lines are bigger than the box
        idx = N.arange(-numsteps, numsteps)
        x = idx*deltax + startx
        y = idx*deltay + starty
        x1 = x - scale*linedx
        x2 = x + scale*linedx
        y1 = y - scale*linedy
        y2 = y + scale*linedy

        # plot lines, clipping to bb
        plotLinesToPainter(painter, x1, y1, x2, y2, bb)

    painter.restore()

# list of fill styles
extfillstyles = (
    'solid', 'horizontal', 'vertical', 'cross',
    'forward diagonals', 'backward diagonals',
    'diagonal cross',
    'forward 2', 'backward 2', 'forward 3',
    'backward 3', 'forward 4', 'backward 4',
    'forward 5', 'backward 5',
    'diagonal cross 2', 'diagonal cross 3',
    'diagonal cross 4', 'diagonal cross 5',
    'vertical forward', 'vertical backward',
    'horizontal forward', 'horizontal backward',
    'star',
    'triangles 1', 'triangles 2', 'triangles 3', 'triangles 4',
    'horizontal double', 'vertical double',
    'forward double', 'backward double',
    'double cross', 'double diagonal cross',
    '94% dense', '88% dense', '63% dense', '50% dense',
    '37% dense', '12% dense', '6% dense',
)

# hatching styles
# map names to lists of (offsetx, offsety, gradx, grady)
_hatchmap = {
    'horizontal': ((0, 0, 0, 1), ),
    'vertical': ((0, 0, 1, 0), ),
    'cross': ( (0, 0, 0, 1), (0, 0, 1, 0), ),
    'forward diagonals': ( (0, 0, 0.7071, -0.7071), ),
    'backward diagonals': ( (0, 0, 0.7071, 0.7071), ),
    'diagonal cross': ( (0, 0, 0.7071, 0.7071), (0, 0, 0.7071, -0.7071), ),
    'forward 2': ( (0, 0, 0.8660, -0.5), ),
    'backward 2': ( (0, 0, 0.8660, 0.5), ),
    'forward 3': ( (0, 0, 0.5, -0.8660), ),
    'backward 3': ( (0, 0, 0.5, 0.8660), ),
    'forward 4': ( (0, 0, 0.2588, -0.9659), ),
    'backward 4': ( (0, 0, 0.2588, 0.9659), ),
    'forward 5': ( (0, 0, 0.9659, -0.2588), ),
    'backward 5': ( (0, 0, 0.9659, 0.2588), ),
    'diagonal cross 2': ( (0, 0, 0.8660, -0.5), (0, 0, 0.8660, 0.5), ),
    'diagonal cross 3': ( (0, 0, 0.5, -0.8660), (0, 0, 0.5, 0.8660), ),
    'diagonal cross 4': ( (0, 0, 0.9659, -0.2588), (0, 0, 0.9659, 0.2588), ),
    'diagonal cross 5': ( (0, 0, 0.2588, 0.9659), (0, 0, 0.2588, -0.9659), ),
    'vertical forward': ( (0, 0, 1, 0), (0, 0, 0.7071, -0.7071), ),
    'vertical backward': ( (0, 0, 1, 0), (0, 0, 0.7071, 0.7071), ),
    'horizontal forward': ( (0, 0, 0, 1), (0, 0, 0.7071, -0.7071), ),
    'horizontal backward': ( (0, 0, 0, 1), (0, 0, 0.7071, 0.7071), ),
    'star': ( (0, 0, 2, 0), (0, 0, 0, 2), (0, 0, -1, 1), (0, 0, 1, 1), ),
    'triangles 1': ( (0, 0, 2, 0), (0, 0, 0, 2), (0, 0, -1, 1),),
    'triangles 2': ( (0, 0, 2, 0), (0, 0, 0, 2), (0, 0, 1, 1),),
    'triangles 3': ( (0, 0, 0, 1), (0, 0, -1, 1), (0, 0, 1, 1), ),
    'triangles 4': ( (0, 0, 1, 0), (0, 0, -1, 1), (0, 0, 1, 1), ),
    'horizontal double': ((0, 0, 0, 1), (0, 0.2828, 0, 1), ),
    'vertical double': ((0, 0, 1, 0), (0.2828, 0, 1, 0), ),
    'forward double': ((0, 0, 0.7071, -0.7071), (0.4,0, 0.7071, -0.7071)),
    'backward double': ((0, 0, 0.7071, 0.7071), (0.4,0, 0.7071, 0.7071)),
    'double cross': (
        (0, 0, 0, 1), (0, 0.2828, 0, 1),
        (0, 0, 1, 0), (0.2828, 0, 1, 0),),
    'double diagonal cross': (
        (0, 0, 0.7071, -0.7071), (0.4,0, 0.7071, -0.7071),
        (0, 0, 0.7071, 0.7071), (0.4,0, 0.7071, 0.7071)),
}

# convert qt-specific fill styles into qt styles
_fillcnvt = {
    'solid': qt.Qt.SolidPattern,
    '94% dense': qt.Qt.Dense1Pattern,
    '88% dense': qt.Qt.Dense2Pattern,
    '63% dense': qt.Qt.Dense3Pattern,
    '50% dense': qt.Qt.Dense4Pattern,
    '37% dense': qt.Qt.Dense5Pattern,
    '12% dense': qt.Qt.Dense6Pattern,
    '6% dense': qt.Qt.Dense7Pattern
}

def brushExtFillPath(painter, extbrush, path, ignorehide=False,
                     stroke=None, dataindex=0):
    """Use an BrushExtended settings object to fill a path on painter.
    If ignorehide is True, ignore the hide setting on the brush object.
    stroke is an optional QPen for stroking outline of path
    """

    if extbrush.hide and not ignorehide:
        if stroke is not None:
            painter.strokePath(path, stroke)
        return

    style = extbrush.style
    if style in _fillcnvt:
        # standard fill: use Qt styles for painting
        color = extbrush.get('color').color(painter, dataindex=dataindex)
        if extbrush.transparency >= 0:
            if extbrush.transparency == 100:
                # skip filling fully transparent areas
                return
            color.setAlphaF((100-extbrush.transparency) / 100.)
        brush = qt.QBrush(color, _fillcnvt[style])
        if stroke is None:
            painter.fillPath(path, brush)
        else:
            painter.save()
            painter.setPen(stroke)
            painter.setBrush(brush)
            painter.drawPath(path)
            painter.restore()

    elif style in _hatchmap:
        # fill with hatching

        if not extbrush.backhide:
            # background brush
            color = extbrush.get('backcolor').color(
                painter, dataindex=dataindex)
            if extbrush.backtransparency > 0:
                color.setAlphaF((100-extbrush.backtransparency) / 100.)
            brush = qt.QBrush(color)
            painter.fillPath(path, brush)

        color = extbrush.get('color').color(painter, dataindex=dataindex)
        if extbrush.transparency >= 0:
            color.setAlphaF((100-extbrush.transparency) / 100.)
        width = extbrush.get('linewidth').convert(painter)
        lstyle, dashpattern = extbrush.get('linestyle')._linecnvt[
            extbrush.linestyle]
        pen = qt.QPen(color, width, lstyle)

        if dashpattern:
            pen.setDashPattern(dashpattern)

        # do hatching with spacing
        spacing = extbrush.get('patternspacing').convert(painter)
        if spacing > 0:
            _hatcher(painter, pen, path, spacing, _hatchmap[style])

        if stroke is not None:
            painter.strokePath(path, stroke)

def brushExtFillPolygon(painter, extbrush, cliprect, polygon, ignorehide=False):
    """Fill a polygon with an extended brush."""
    clipped = qt.QPolygonF()
    polygonClip(polygon, cliprect, clipped)
    path = qt.QPainterPath()
    path.addPolygon(clipped)
    brushExtFillPath(painter, extbrush, path, ignorehide=ignorehide)
